@@ -30,8 +30,12 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <sys/types.h>
 #include <unistd.h>
+#include <fcntl.h>
+#include <errno.h>
 #include <time.h>
+#include <poll.h>
 
 #ifdef HAVE_ESD
 #include <esd.h>
@@ -40,6 +44,7 @@
 #ifdef HAVE_ESD
 static char *esound_hostname = NULL;
 static int gnome_sound_connection = -1;
+static gboolean gnome_sound_enabled = TRUE;
 #endif
 
 typedef struct _sample
@@ -297,18 +302,39 @@ gnome_sound_sample_load_wav(const char *file)
 static gboolean
 use_sound (void)
 {
+  if (!gnome_sound_enabled)
+    return FALSE;
+
   if (gnome_sound_connection == -1){
       gnome_sound_connection = esd_open_sound (esound_hostname);
-      if (gnome_sound_connection == -1){
-	g_free (esound_hostname);
-	esound_hostname = NULL;
-	return FALSE;
-      }
-    }
-  return TRUE;
+  }
+
+  return gnome_sound_connection >= 0;
+}
+
+static void
+close_sound_connection (void)
+{
+  if(gnome_sound_connection >= 0){
+    esd_close(gnome_sound_connection);
+    gnome_sound_connection = -1;
+  }
 }
 #endif
 
+G_GNUC_INTERNAL void _gnome_sound_set_enabled (gboolean);
+
+void G_GNUC_INTERNAL
+_gnome_sound_set_enabled (gboolean enabled)
+{
+#ifdef HAVE_ESD
+  gnome_sound_enabled = enabled;
+  if (!enabled) {
+    close_sound_connection();
+  }
+#endif
+}
+
 #if defined(HAVE_LIBAUDIOFILE) && defined(HAVE_ESD)
 #include <audiofile.h>
 
@@ -391,6 +417,67 @@ gnome_sound_sample_load_audiofile(const char *file)
 }
 #endif
 
+#ifdef HAVE_ESD
+static int
+send_all (int fd, const char *buf, size_t buflen)
+{
+	struct pollfd pfd[1];
+	size_t nwritten = 0;
+	int flags, rv;
+	ssize_t n;
+	
+	if ((flags = fcntl (fd, F_GETFL)) == -1)
+		return -1;
+	
+	fcntl (fd, F_SETFL, flags | O_NONBLOCK);
+	
+	pfd[0].events = POLLOUT;
+	pfd[0].fd = fd;
+	
+	do {
+		do {
+			pfd[0].revents = 0;
+			rv = poll (pfd, 1, 100);
+		} while (rv == -1 && (errno == EINTR || errno == EAGAIN));
+		
+		if (rv == -1) {
+			if (errno != EBADF) {
+				rv = errno;
+				esd_close (fd);
+				errno = rv;
+			}
+			
+			fd = -1;
+			break;
+		} else if (rv < 1 || (pfd[0].revents & (POLLERR | POLLHUP | POLLOUT)) != POLLOUT) {
+			/* we /just/ lost the esd connection */
+			esd_close (fd);
+			fd = -1;
+			break;
+		} else {
+			/* socket is ready for writing */
+			do {
+				n = write (fd, buf + nwritten, buflen - nwritten);
+			} while (n == -1 && errno == EINTR);
+			
+			if (n == -1) {
+				rv = errno;
+				esd_close (fd);
+				errno = rv;
+				fd = -1;
+				break;
+			}
+			
+			nwritten += n;
+		}
+	} while (nwritten < buflen);
+	
+	if (fd != -1 && flags != -1)
+		fcntl (fd, F_SETFL, flags);
+	
+	return fd;
+}
+#endif
 
 /**
  * gnome_sound_sample_load:
@@ -447,21 +534,23 @@ gnome_sound_sample_load(const char *sample_name, const char *filename)
 	   * file, or event type, for later identification */
 	  s->id = esd_sample_cache (gnome_sound_connection, s->format, s->rate,
 				    size, (char *)sample_name);
-	  write (gnome_sound_connection, s->data, size);
-	  confirm = esd_confirm_sample_cache (gnome_sound_connection);
+	  
+	  gnome_sound_connection = send_all (gnome_sound_connection, (const char *) s->data, size);
+	  if (gnome_sound_connection != -1)
+	    confirm = esd_confirm_sample_cache (gnome_sound_connection);
+	  
 	  if (s->id <= 0 || confirm != s->id)
 	    {
 	      g_warning ("error caching sample <%d>!\n", s->id);
 	      s->id = 0;
 	    }
-	  g_free (s->data);
-	  s->data = NULL;
 	}
     }
 
   sample_id = s->id;
 
-  g_free(s->data); g_free(s);
+  g_free(s->data);
+  g_free(s);
 
   return sample_id;
 #else
@@ -499,9 +588,12 @@ gnome_sound_play (const char * filename)
 
   sample = gnome_sound_sample_load (buf, filename);
 
-  esd_sample_play(gnome_sound_connection, sample);
-  fsync (gnome_sound_connection);
-  esd_sample_free(gnome_sound_connection, sample);
+  if (gnome_sound_connection != -1 && sample > 0)
+    {
+      esd_sample_play(gnome_sound_connection, sample);
+      fsync (gnome_sound_connection);
+      esd_sample_free(gnome_sound_connection, sample);
+    }
 #endif
 }
 
@@ -517,9 +609,7 @@ gnome_sound_init(const char *hostname)
 #ifdef HAVE_ESD
 	srand(time(NULL));
 	g_free (esound_hostname);
-	esound_hostname = NULL;
-	if (hostname)
-		esound_hostname = g_strdup (hostname);
+	esound_hostname = g_strdup (hostname);
 #endif
 }
 
@@ -535,10 +625,7 @@ gnome_sound_shutdown(void)
 	g_free (esound_hostname);
 	esound_hostname = NULL;
 
-	if(gnome_sound_connection >= 0){
-		esd_close(gnome_sound_connection);
-		gnome_sound_connection = -1;
-	}
+        close_sound_connection ();
 #endif
 }
 
